import sys
import os
import hashlib
import json
from pathlib import Path

from .environment import (
    get_package_directory,
    WINDOWS,
    LINUX,
    MACOS,
    short_envname,
    get_scripts_dir,
    get_distribution_of_module,
    detect_conda_env,
)
from .windows import (
    get_start_menu,
    create_shortcut,
    set_process_appusermodel_id,
    unredirect_appdata,
    refresh_shell_cache,
)
from .fix_entry_points import fix_entry_points
from .linux import get_user_applications, create_desktop_file


CONFIG_FILENAME = 'desktop-app.json'


class _ModuleConfig:
    """Object holding desktop-app configuration for a module"""

    _instances = {}

    @classmethod
    def instance(cls, module_name):
        """Get an existing instance, if any, or create one and return it."""
        if module_name not in cls._instances:
            cls._instances[module_name] = cls(module_name)
        return cls._instances[module_name]

    def __init__(self, module_name):
        self.module_name = module_name
        self.package_directory = get_package_directory(module_name)
        self.module_directory = Path(
            self.package_directory, *self.module_name.split('.')[1:]
        )

        config = self._load_config()

        self.org_name = config.get('org_name', None)
        module_config = config.get('modules', {}).get(module_name, {})

        self.display_name = module_config.get('display_name', module_name)
        env = short_envname()
        if env is not None:
            self.display_name += f' ({env})'

        winicon = module_config.get('winicon', None)
        if winicon is None:
            self.winicon = self.module_directory / f'{module_name}.ico'
        else:
            self.winicon = self.package_directory / winicon

        icon = module_config.get('icon', None)
        if icon is None:
            self.icon = self.module_directory / f'{module_name}.svg'
            if not self.icon.exists():
                self.icon = self.module_directory / f'{module_name}.png'
        else:
            self.icon = self.package_directory / icon

        self.launcher_script_path = self._get_launcher_script_path()
        self.appid = self._get_appid()

    def _load_config(self):
        """Load 'desktop-app.json' from the module package directory and return it. If
        it doesn't exist, return an empty dict."""
        try:
            return json.loads(
                (self.package_directory / CONFIG_FILENAME).read_text(encoding='utf8')
            )
        except FileNotFoundError:
            return {}

    def _get_launcher_script_path(self):
        """Get the path to the script for launching the app without a console. It is
        assumed to be called <module_name>-gui.exe on Windows and <module-name>
        otherwise, and to be in the bin or Scripts directory corresponding to the
        installation location of the module. If the module is not installed, returns
        None."""
        # Look up the path to the launcher:
        scripts_dir = get_scripts_dir(self.module_name)
        if scripts_dir is None:
            return None
        script_path = scripts_dir / self.module_name
        if WINDOWS:
            return Path(script_path.parent, script_path.name + '-gui.exe')
        return script_path

    def _get_appid(self):
        """ Create a string identifying the application, for use in OS interfaces such
        as as a Windows AppUserModelID. on Windows we use:

        <OrgName>.<ModuleName>.Python-<hexdigits>

        OrgName is omitted if not present in the configuration. The last field contains
        a hash of the path to the Python interpreter for the module, ensuring the appid
        is unique to the Python environment. The first three fields are converted to
        CamelCase, periods replaced with hyphens and underscores removed.

        On Linux we use:

        <module_name>-<envname>

        where <envname> is the name of the conda or virtual environment. The env is
        omitted if it's a conda env called 'base' or a virtual environment called
        '.venv' or 'venv'. The reason we use a different format to Windows is that this
        will be used as the executable name of the app's process (sys.argv[0]), and so
        it's slightly more important for it to be user-friendly as it will appear as the
        default window title in some GUI toolkits.

        On macos it's the same as Linux, but this is likely to change when proper macos
        support is introduced.

        """
        # Hash the install prefix so that we can include in the appid a segment unique
        # to the Python environment. Normalise the case first, since I've observed that
        # in within a venv sys.executable is all lower case.
        if WINDOWS:
            replacements = {' ': '', '_': '', '.': '-'}
            install_prefix = Path(os.path.normcase(sys.prefix))
            interpreter_hash = hashlib.sha256(bytes(install_prefix)).hexdigest()[:16]
            appid_parts = []
            for part in [self.org_name, self.module_name]:
                if part is None:
                    continue
                camelcase_part = part.title()
                for a, b in replacements.items():
                    camelcase_part = camelcase_part.replace(a, b)
                appid_parts.append(camelcase_part)
            appid_parts.append(f'Python-{interpreter_hash}')
            return '.'.join(appid_parts)
        else:  # TODO: consider macos
            env = short_envname()
            if env is None:
                return self.module_name
            return f'{self.module_name}-{env}'


def set_process_appid(module_name):
    """Associate the currently running process with the shortcut for the given module
    name. This should ensure the app has the correct icon in the taskbar, groups its
    windows correctly, can be pinned etc."""

    config = _ModuleConfig.instance(module_name)
    launcher = config.launcher_script_path
    if WINDOWS:
        set_process_appusermodel_id(config.appid)
    else:
        # Most Linux GUI toolkits set the X WM_CLASS property from the basename of
        # sys.argv[0], so ensuring it matches the name of our .desktop file is
        # sufficient to get DEs to correctly identify our app windows. Hopefully the
        # toolkits do the equivalent in Wayland - setting the app_id xdg-shell property.
        # If not, the user will need to make the right function call depending on their
        # GUI toolkit. Notable exception: tk doesn't use sys.argv[0] - the user needs to
        # set the tk classname to sys.argv[0] themselves.
        launcher = config.launcher_script_path
        # If the launcher script path is None, that means the package isn't installed
        # into (user-)site-packages, so the script doesn't exist. Don't set sys.argv[0].
        # We don't want to set ays.argv to something that could not actually be run
        # again.
        if launcher is not None and launcher.exists():
            symlink_path = _launcher_script_symlink_path(config)
            if symlink_path is not None and symlink_path.exists():
                sys.argv[0] = str(symlink_path)
            else:
                sys.argv[0] = str(config.launcher_script_path)
    # TODO: consider macos


def _default_shortcut_dir(config):
    if WINDOWS:
        path = get_start_menu()
        if config.org_name is not None:
            return path / config.org_name
        return path
    elif LINUX:
        return get_user_applications()
    elif MACOS:
        raise NotImplementedError


def _shortcut_name(config):
    if WINDOWS:
        return f'{config.display_name}.lnk'
    elif LINUX:
        return f'{config.appid}.desktop'
    elif MACOS:
        raise NotImplementedError


def _launcher_script_symlink_path(config):
    if not WINDOWS and config.appid != config.module_name:
        if config.launcher_script_path is not None:
            return config.launcher_script_path.parent / config.appid


def install(module_name, path=None, verbose=False, no_fix_entry_points=False):
    """Add a shortcut to launch the app for the given module to the system menus, i.e
    the start menu on Windows, ~/.local/share/applications on Linux, and TODO on macOS.
    If path is given, the shortcut will be created in the given directory instead. On
    Linux, in order to ensure the name of the shortcut (which may contain the name of a
    conda or virtual environment) matches the name of the executable it points to, a
    symbolic link called `<module_name>-<envname>` will be created."""
    config = _ModuleConfig.instance(module_name)
    if config.launcher_script_path is None:
        msg = f"""The package providing the module {module_name} is not installed to the
            current Python environment, so its entry_points scripts do not exist, and
            desktop-app cannot create shortcuts to them. Install the package to the
            Python environment before continuing. If you are in a development setup you
            may make an editable install with `pip install -e`."""
        raise EnvironmentError(' '.join(msg.split()))
    if path is not None:
        path = Path(path)
    else:
        path = _default_shortcut_dir(config)
    shortcut_path = path / _shortcut_name(config)
    if shortcut_path.exists():
        if verbose:
            msg = f'warning: overwriting existing file {shortcut_path}'
            print(msg, file=sys.stderr)
        shortcut_path.unlink()
    if WINDOWS:
        create_shortcut(
            shortcut_path,
            config.launcher_script_path,
            working_directory=Path.home(),
            icon_file=config.winicon,
            display_name=config.display_name,
            appusermodel_id=config.appid,
        )
        # If the shortcut is in appdata, and if it's a private copy of appdata such that
        # it will not be used to populate the start menu, move it to real appdata where
        # it will be used to populate the start menu. Hopefully in time Windows changes
        # to consider the start menus within these private directories as places that it
        # populates the start menu from. If not, hopefully this workaround continues to
        # function.
        unredirect_appdata(shortcut_path)
        refresh_shell_cache()
        if verbose:
            print(f' -> created {shortcut_path}')
        conda_env, _ = detect_conda_env()
        if conda_env is not None:
            # Re-create entry points for the package, to work around the fact that conda
            # does not support gui_scripts, so those entry_points may have been
            # installed as console_scripts:
            dist = get_distribution_of_module(module_name)
            if not no_fix_entry_points:
                fix_entry_points(dist)
    elif LINUX:
        symlink_path = _launcher_script_symlink_path(config)
        create_desktop_file(
            shortcut_path,
            target=symlink_path or config.launcher_script_path,
            display_name=config.display_name,
            icon_file=config.icon,
        )
        if verbose:
            print(f' -> created {shortcut_path}')
        if symlink_path is not None:
            if symlink_path.exists():
                if verbose:
                    msg = f'warning: overwriting existing symlink {symlink_path}'
                    print(msg, file=sys.stderr)
                    symlink_path.unlink()
            symlink_path.symlink_to(config.launcher_script_path)
            if verbose:
                print(
                    f' -> created symlink {symlink_path} -> '
                    + f'{config.launcher_script_path.name}'
                )
    elif MACOS:
        raise NotImplementedError


def uninstall(module_name, path=None, verbose=False):
    """Remove the app shortcut for the given module from the system menus, i.e the start
    menu on Windows, ~/.local/share/applications on Linux, and TODO on macOS. If a
    symlink was created within the scripts folder in install(), delete it."""
    config = _ModuleConfig.instance(module_name)
    if path is not None:
        path = Path(path)
    else:
        path = _default_shortcut_dir(config)
    files_to_delete = [path / _shortcut_name(config)]
    symlink_path = _launcher_script_symlink_path(config)
    if symlink_path is not None:
        files_to_delete.append(symlink_path)
    for file in files_to_delete:
        try:
            file.unlink()
            if verbose:
                print(f' -> deleted {file}')
        except FileNotFoundError:
            if verbose:
                print(f'warning: no such file {file}', file=sys.stderr)
    if WINDOWS:
        refresh_shell_cache()



from .__version__ import __version__
from .launcher import entry_point
from .shell import set_process_appid, install, uninstall

__all__ = [
    'entry_point',
    'set_process_appid',
    'install',
    'uninstall',
]



import site
from pathlib import Path
from subprocess import list2cmdline, call, Popen, PIPE, DEVNULL
from .environment import WINDOWS


if WINDOWS:
    # Allow importing this module on other platforms even though the functions will fail
    from win32com.shell import shell, shellcon
    from win32com.client import Dispatch
    from win32com.propsys import propsys, pscon
    from .wineventhook import sethook


def _checkwindows():
    if not WINDOWS:
        msg = "Windows-only function called on a non-Windows OS"
        raise EnvironmentError(msg)


def get_start_menu():
    """Return the path for Start menu entries for user-installed programs"""
    _checkwindows()
    return Path(Dispatch('WScript.Shell').SpecialFolders("Programs"))


def unredirect_appdata(path):
    """Copy a file within the appdata directory, that is actually in a local private
    copy of said directory, into the regular, shared-by-all-apps appdata directory.
    Delete the original file (but not necessarily parent directories). This is necessary
    if running on Python from the Windows Store, since Python.exe sees a private copy of
    appdata."""
    # The way this hack works is that we look site site.USER_BASE. This is the only clue
    # we have that our appdata is being redirected, since if it is, site.USER_BASE will
    # be in the redirected location. So that's how we get our hands on the path (this
    # ultimately came into the Python interpreter via an environment variable). Then we
    # launch non-python subprocesses to do the file copying for us, since the
    # subprocesses aren't subject to the redirect.
    real_path = Path(path)
    real_appdata = Path(Dispatch('WScript.Shell').SpecialFolders("AppData"))
    sandbox_appdata = Path(site.USER_BASE).parent / real_appdata.name
    sandbox_path = sandbox_appdata / real_path.relative_to(real_appdata)
    if sandbox_path.exists():
        # Yep, the file was redirected to the private copy of appdata:
        call(
            ['cmd.exe', '/c', 'mkdir', str(real_path.parent)],
            stdout=DEVNULL,
            stderr=DEVNULL,
        )
        proc = Popen(
            ['xcopy', str(sandbox_path), str(real_path), '/Y'],
            stdin=PIPE,
            stdout=DEVNULL,
            stderr=DEVNULL,
        )
        proc.communicate(b'F')  # This is so dumb
        sandbox_path.unlink()


def create_shortcut(
    shortcut_path,
    target,
    arguments=(),
    working_directory=None,
    icon_file=None,
    display_name=None,
    appusermodel_id=None,
):
    """Create a Windows shortcut at the given path, which should be a filepath ending in
    '.lnk'. Arguments should be a list or tuple of arguments - not a single string of
    arguments separated by spaces."""
    shortcut_path = Path(shortcut_path)
    shortcut_path.parent.mkdir(parents=True, exist_ok=True)
    _checkwindows()
    objShell = Dispatch('WScript.Shell')
    shortcut = objShell.CreateShortcut(str(shortcut_path))
    shortcut.TargetPath = str(target)
    if arguments:
        shortcut.Arguments = list2cmdline(arguments)
    if working_directory is not None:
        shortcut.WorkingDirectory = str(working_directory)
    if icon_file is not None:
        shortcut.IconLocation = str(icon_file)
    if display_name is not None:
        shortcut.Description = display_name
    shortcut.save()

    if appusermodel_id is not None:
        # Edit the shortcut to associate the AppUserModel_ID with it:
        store = propsys.SHGetPropertyStoreFromParsingName(
            str(shortcut_path), None, shellcon.GPS_READWRITE, propsys.IID_IPropertyStore
        )
        store.SetValue(
            pscon.PKEY_AppUserModel_ID, propsys.PROPVARIANTType(appusermodel_id)
        )
        store.Commit()


def refresh_shell_cache():
    # Refresh the icon cache:
    shell.SHChangeNotify(
        shellcon.SHCNE_ASSOCCHANGED,
        shellcon.SHCNF_IDLIST | shellcon.SHCNF_FLUSH,
        None,
        None,
    )


def set_process_appusermodel_id(appid):
    _checkwindows()
    # shell.SetCurrentProcessExplicitAppUserModelID(appid)
    if not isinstance(appid, str) or len(appid) >= 1024:
        raise TypeError('appid must be a str of len < 1024')
    sethook(appid)



from setuptools import setup

setup(
    name='oink',
    version='1.0',
    author='Old MacDonald',
    author_email="macdonald@eie.io",
    packages=["oink"],
    setup_requires=['setuptools'],
    install_requires=["desktop-app"],
    zip_safe=False,
    include_package_data=True,
    package_data={"": ['*.svg', '*.ico', 'desktop-app.json']},
    entry_points={
        'console_scripts': ['oink = desktop_app:entry_point'],
        'gui_scripts': ['oink-gui = desktop_app:entry_point'],
    },
)



import sys
import os
import site
import sysconfig
from importlib.util import find_spec
import platform
from functools import lru_cache
from pathlib import Path
try:
    import importlib.metadata as importlib_metadata
except ImportError:
    import importlib_metadata

_os = platform.system()
WINDOWS = _os == 'Windows'
LINUX = _os == 'Linux'
MACOS = _os == 'Darwin'
if not (WINDOWS or LINUX or MACOS):
    raise EnvironmentError("Can't determine operating system")


# Three possible locations for scripts.

# Main scripts dir for the Python environment:
PY_SCRIPTS = Path(sysconfig.get_path('scripts'))
# Used in Debian-based distros for packages installed with `sudo pip install`:
LOCAL_SCRIPTS = Path('/usr/local/bin')
# Packages installed to the user's home directory with `pip install --local`. Even if
# the user doesn't type `--local`, this option is enabled by default on Debian-based
# distros when running without sudo, and for the Python shipped in the Microsoft Store:
if WINDOWS:
    USER_SCRIPTS = Path(sysconfig.get_path('scripts', scheme='nt_user'))
else:
    USER_SCRIPTS = Path(sysconfig.get_path('scripts', scheme='posix_user'))

# Possible locations for installed packages:

# Arbitrary number of site packages directories:
def _site_packages():
    return [Path(s) for s in site.getsitepackages()]


# Only one user site packages directory:
def _user_site_packages():
    return Path(site.getusersitepackages())


def _reverse_egg_link_lookup(directory):
    # For packages 'installed' with `pip install -e`, the import path is the development
    # directory, which is not very useful in figuring out what the corresponding scripts
    # directory is. This function looks through canonical installation locations for
    # .egg-link files, and if any is found that points to the given directory, the
    # parent directory of the .egg-link file is returned. Otherwise returns None.
    directory = Path(directory).absolute()
    for sitedir in _site_packages() + [_user_site_packages()]:
        if not sitedir.is_dir():
            continue
        sitedir = Path(sitedir).absolute()
        if sitedir.exists and sitedir.is_dir:
            for file in sitedir.iterdir():
                if file.suffix == '.egg-link':
                    # The first line is the path to the .egg:
                    linkpath = Path(file.read_text("utf8").splitlines()[0])
                    # is allowed to be relative to the containing dir:
                    linkpath = Path(sitedir, linkpath)
                    if linkpath == directory.parent:
                        return sitedir

    # For new PEP660-style editable installs, we instead look a) through sys.modules for
    # modules with names starting with '__editable__', and inspect their MAPPING
    # attribute, and b) through site_packages for __editable__*.pth files specifying a
    # simple
    for name, module in sys.modules.items():
        if name.startswith('__editable__'):
            if directory in [Path(s) for s in module.MAPPING.values()]:
                return Path(module.__file__).parent

    # And now for src-layout packages.

    # Bit of duplication here with the above code for .egg-link files. Prefer this to
    # keep it separate.
    for sitedir in _site_packages() + [_user_site_packages()]:
        if not sitedir.is_dir():
            continue
        sitedir = Path(sitedir).absolute()
        if sitedir.exists and sitedir.is_dir:
            for file in sitedir.iterdir():
                if file.name.startswith('__editable__') and file.suffix == '.pth':
                    contents = file.read_text("utf8")
                    if contents.startswith("import"):
                        # This one's a dynamic loader, skip it.
                        continue
                    # The first line is the path to the .egg:
                    linkpath = contents.strip()
                    # is allowed to be relative to the containing dir:
                    linkpath = Path(sitedir, linkpath)
                    if linkpath == directory.parent:
                        return sitedir


def get_install_directory(module_name):
    """Return the installation directory of the module - an entry in the list of site
    packages directories, or the user site packages directory."""
    import_path = get_package_directory(module_name)
    for sitedir in _site_packages() + [_user_site_packages()]:
        if sitedir in import_path.parents:
            return sitedir
    # May return None if there is no egg-link to the package directory either
    return _reverse_egg_link_lookup(import_path)


def get_scripts_dir(module_name):
    """Return the directory of the scripts installed with the package supplying the
    given module. If the package is not installed, returns None."""
    install_dir = get_install_directory(module_name)
    if install_dir is not None:
        if Path('/usr/local') in install_dir.parents:
            return LOCAL_SCRIPTS
        elif install_dir in _site_packages():
            return PY_SCRIPTS
        elif install_dir == _user_site_packages():
            return USER_SCRIPTS


def get_package_directory(module_name):
    """Return Path of the package directory that the given module is in."""
    base_module = module_name.split('.', 1)[0]
    spec = find_spec(base_module)
    if spec is None or spec.origin is None:
        raise ModuleNotFoundError(base_module)
    if not spec.parent:
        msg = f'{base_module} is not a package'
        raise ValueError(msg)
    return Path(spec.origin).parent


def get_distribution_of_module(module_name):
    """Return the name of the distribution providing the given module"""
    base_module = module_name.split('.', 1)[0]
    for distribution in importlib_metadata.distributions():
        if distribution.files is None:
            continue
        for path in distribution.files:
            p = Path(path).parts[0]
            if (p == base_module or # standard installation or pre-PEP660 editable
                p.startswith(f'__editable__.{base_module:s}')): # PEP660 editable

                return distribution.metadata['Name']


@lru_cache()
def detect_conda_env():
    """Inspect whether `sys.executable` is within a conda environment and if it is,
    return the environment name and Path of its prefix. Otherwise return None, None"""
    prefix = Path(sys.prefix)
    if not (prefix / 'conda-meta').is_dir():
        # Not a conda env
        return None, None
    if (prefix / 'envs').is_dir():
        # It's the base conda env:
        return 'base', prefix
    # Not the base env: its name is the directory basename:
    return prefix.name, prefix


def activate_conda_env(name, prefix):
    """Modify environment variables so as to effectively activate the given conda env
    from the perspective of child processes. If the conda env appears to already be
    active, do nothing. Does not actually set environment variables, instead returns a
    copy that may be passed to subprocess.Popen as the env arg. Not all environment
    variables are considered, only CONDA_DEFAULT_ENV, CONDA_PREFIX, and PATH."""
    prefix = Path(prefix)
    env = os.environ.copy()
    current_name = env.get('CONDA_DEFAULT_ENV')
    current_prefix = env.get('CONDA_PREFIX')
    if current_prefix is not None:
        current_prefix = Path(current_prefix)
    if current_name == name and current_prefix == prefix:
        # Env is already active
        return
    env['CONDA_DEFAULT_ENV'] = name
    env['CONDA_PREFIX'] = str(prefix)
    if WINDOWS:
        env['PATH'] = os.path.pathsep.join(
            [
                str(prefix),
                str(prefix / "Library" / "mingw-w64" / "bin"),
                str(prefix / "Library" / "usr" / "bin"),
                str(prefix / "Library" / "bin"),
                str(prefix / "Scripts"),
                env['PATH'],
            ]
        )
    else:
        env['PATH'] = os.path.pathsep.join([str(prefix / 'bin'), env['PATH']])

    return env


@lru_cache()
def detect_venv():
    """Inspect whether sys.executable is within a virtual environment and if it is,
    return the virtual environment Path of prefix. Otherwise return None"""
    if hasattr(sys, 'real_prefix'):
        # virtualenv < v20 sets sys.real_prefix, which doesn't exist otherwise
        return Path(sys.prefix)
    if sys.base_prefix != sys.prefix:
        # virtualenv >= v20 sets sys.base_prefix, which always exists
        return Path(sys.prefix)


def activate_venv(prefix):
    """Modify environment variables so as to effectively activate the given virtual
    environment from the perspective of child processes. If the virtual environment
    appears to already be active, do nothing. Does not actually set environment
    variables, instead returns a copy that may be passed to subprocess.Popen as the env
    arg."""
    prefix = Path(prefix)
    env = os.environ.copy()
    current_env = env.get('VIRTUAL_ENV')
    if current_env is not None:
        current_env = Path(current_env)
    if current_env == prefix:
        # Env is already active
        return
    env['VIRTUAL_ENV'] = str(prefix)
    env.pop('PYTHONHOME', None)
    if WINDOWS:
        env['PATH'] = os.pathsep.join([str(prefix / 'Scripts'), env['PATH']])
    else:
        env['PATH'] = os.pathsep.join([str(prefix / 'bin'), env['PATH']])
    return env


def short_envname():
    """Get a short name useful for distinguishing an environment from other
    environments. For conda, it's the name of the env or None if it's the base env. For
    venv its the name of the venv directory with any leading '.' stripped off, or None
    if that results in 'venv'. It's also None if we're not in either kind of
    environment"""
    envname, _ = detect_conda_env()
    if envname is not None:
        if envname == 'base':
            return None
        return envname
    envpath = detect_venv()
    if envpath is not None:
        envname = envpath.name.lstrip('.')
        if envname == 'venv':
            return None
        return envname

def get_venv_executable_and_env(env=None):
    """If we are running in a venv on Windows or MacOS such that sys.executable is a
    shim to sys._base_executable, then return sys._base_executable, as well as the
    environment that should be passed to sys._base_executable in order for it to know to
    configure itself for the virtual environment instead of the base environment. This
    allows us to skip the shim and thus have fewer layers in our process hierarchy. If
    we are not in a venv, return sys.executable and the env as passed in. If env passed
    in is None, and needs to be modified for the subprocess, a copy of os.environ(),
    appropriately modified, will be returned."""
    base_executable = getattr(sys, '_base_executable', sys.executable)
    if Path(base_executable) != Path(sys.executable):
        if env is None:
            env = os.environ.copy()
        env["__PYVENV_LAUNCHER__"] = str(sys.executable)
        return base_executable, env
    return sys.executable, env




import sys
import tkinter
import desktop_app

desktop_app.set_process_appid("oink")


def main():
    root = tkinter.Tk(className=sys.argv[0])
    root.geometry("500x300")
    w = tkinter.Label(root, text="Oink!")
    w.place(relx=0.5, rely=0.3, anchor=tkinter.CENTER)
    sys_info = (
        f"sys.prefix: {sys.prefix}\n"
        f"sys.exec_prefix: {sys.exec_prefix}\n"
        f"sys.executable: {sys.executable}"
    )
    s = tkinter.Label(root, text=sys_info, justify=tkinter.LEFT)
    s.place(relx=0.05, rely=0.5, anchor=tkinter.W)
    print(sys_info)
    root.mainloop()


if __name__ == "__main__":
    main()



import sys
import subprocess
from pathlib import Path

from .environment import (
    detect_conda_env,
    activate_conda_env,
    detect_venv,
    activate_venv,
    get_package_directory,
    WINDOWS,
    get_venv_executable_and_env,
)


def entry_point():
    """Run a child Python process, passing it `[<script>] + sys.argv[1:]` as arguments.
    When complete, `sys.exit()` with its return code.

    `<script>` is determined from `sys.argv[0]`, which is assumed to be an
    `entry_points` script named as a dotted module path. A suffix '-gui', if present, is
    stripped from `sys.argv[0]` before treating it as a module name. `<script>` is set
    to the filepath of that module, or its `__main__.py` if it is a package.

    If `sys.executable` is Pythonw.exe, the subprocess is run with the corresponding
    Python.exe, but with `creationflags` set to `CREATE_NO_WINDOW` such that the process
    is started with a hidden console window. Otherwise the subprocess is run with
    `sys.executable`.

    If `sys.executable` appears to be within a conda env or virtual environment, then
    the environment variables of the child process will be modified such that the
    environment is effectively activated for the subprocess. In this way an
    `entry_points` script may be run directly (such as from a start menu shortcut)
    without activating the virtual environment it is contained within, and the child
    process will still see an activated environment.

    Note that the path to the module is passed directly to Python as a script. This
    differs to running `python -m <modulename>` in that if the module is a package, its
    `__init__.py` will not be run first. This is something of an optimisation to allow
    GUI programs to display a splash screen before doing any other imports. If important
    initialisation is done in `__init__`, then the script should import the main package
    before such initialisation is required.
    """
    # Resolve symlinks on Unix since we actually create symlinks to the script - but
    # dont' call resolve() on Windows since this can raise an error for Microsoft Store
    # Python, where we don't have permission to resolve that path (even though we're
    # allowed to run it!)
    if WINDOWS:
        script = Path(sys.argv[0])
    else:
        script = Path(sys.argv[0]).resolve()

    module_name, *_ = script.name.rsplit('-gui', 1)
    # Find the path of the module:
    package_directory = get_package_directory(module_name)
    script_path = Path(package_directory, *module_name.split('.')[1:])
    if script_path.is_dir():
        script_path /= '__main__.py'
    else:
        script_path = script_path.with_suffix('.py')

    popen_kwargs = {}

    conda_envname, conda_prefix = detect_conda_env()
    if conda_envname is not None:
        env = activate_conda_env(conda_envname, conda_prefix)
        popen_kwargs['env'] = env

    venv_prefix = detect_venv()
    if venv_prefix is not None:
        env = activate_venv(venv_prefix)
        popen_kwargs['env'] = env

    python, popen_kwargs['env'] = get_venv_executable_and_env(popen_kwargs.get('env'))
    python = Path(python)
    # Case-insensitive comparison requried on Windows
    if python.name.lower() == 'pythonw.exe':
        python = python.parent / 'python.exe'
        # TODO: can use subprocess.CREATE_NO_WINDOW once Python 3.6 is end-of-life
        CREATE_NO_WINDOW = 1 << 27
        popen_kwargs['creationflags'] = CREATE_NO_WINDOW

    try:
        sys.exit(
            subprocess.call(
                [str(python), str(script_path)] + sys.argv[1:], **popen_kwargs
            )
        )
    except KeyboardInterrupt:
        sys.exit(1)



import argparse
from .shell import install, uninstall


def main():
    parser = argparse.ArgumentParser(
        description="""Create (or remove) a Start menu shortcut (Windows) or .desktop
            file (Linux) to run the Python module of the given name. The package owning
            the module must have configured appropriate entry_points for the module, and
            either have a desktop-app.json specifying the location of the icon files to
            use, or must have the files in the default locations. See the main
            desktop-app documentation for details."""
    )

    parser.add_argument(
        action="store",
        choices=['install', 'uninstall'],
        dest="action",
        help="""Whether to create (install) a shortcut or .desktop file, or to delete
            (uninstall) an existing one.""",
    )
    parser.add_argument(
        '--path',
        action="store",
        default=None,
        help="""Directory to create/delete shortcut or .desktop file. If not given,
            defaults to the Start menu on Windows, and to ~/.local/share/applications on
            Linux.""",
    )

    parser.add_argument(
        '--no-fix-entry-points',
        dest='no_fix_entry_points',
        default=False,
        action='store_true',
        help="If set, entry points will not be fixed for conda-environments "
        "during the install action."
    )
    parser.add_argument(
        '-q',
        '--quiet',
        action="store_true",
        help="""Don't print the names of files created/deleted.""",
    )
    parser.add_argument(action="store", dest="module", nargs='+')

    args = parser.parse_args()
    if args.action == 'install':
        for module in args.module:
            install(module, path=args.path, verbose=not args.quiet,
                    no_fix_entry_points=args.no_fix_entry_points)
    if args.action == 'uninstall':
        for module in args.module:
            uninstall(module, path=args.path, verbose=not args.quiet)


if __name__ == '__main__':
    main()



import sys
import sysconfig
try:
    import importlib.metadata as importlib_metadata
except ImportError:
    import importlib_metadata
from distlib.scripts import ScriptMaker


def fix_entry_points(*distribution_names):
    """Re-make all entry_points for the given distributions. This is a workaround for
    the fact that conda doesn't support GUI scripts. The filenames on disk remain the
    same, so conda can still uninstall cleanly - we're just fixing the files in-place"""

    # There are many places scripts can go, but we're in conda so the main scripts dir
    # at sysconfig.get_path('scripts') is the only possibility for us:
    maker = ScriptMaker(None, sysconfig.get_path('scripts'))
    maker.clobber = True  # Overwrite existing scripts
    maker.variants = {''}  # Don't make variants with Python major.minor suffixes

    for name in distribution_names:
        # Get the script specs and format them how Scriptmaker wants them
        distribution = importlib_metadata.Distribution.from_name(name)

        # I've seen enough bugs with entry_points in conda that I feel the need to
        # clobber its console_scripts too, even though they are officially supported:
        console_scripts = [
            f'{e.name} = {e.value}'
            for e in distribution.entry_points
            if e.group == 'console_scripts'
        ]
        gui_scripts = [
            f'{e.name} = {e.value}'
            for e in distribution.entry_points
            if e.group == 'gui_scripts'
        ]

        # Make 'em:
        maker.make_multiple(console_scripts)
        maker.make_multiple(gui_scripts, {'gui': True})


if __name__ == '__main__':
    fix_entry_points(*sys.argv[1:])


import os
from textwrap import dedent
from pathlib import Path


def _desktop_escape(s):
    """Escape a filepath for use in a .desktop file"""
    escapes = {' ': R'\s', '\n': R'\n', '\t': R'\t', '\\': R'\\'}
    s = str(s)
    for unescaped, escaped in escapes.items():
        s = s.replace(unescaped, escaped)
    return s


def get_user_applications():
    """Return the path to the user applications directory, where user .desktop files
    should be placed, typically ~/local/share/applications."""
    data_home = os.getenv('XDG_DATA_HOME')
    if data_home is not None:
        data_home = Path(data_home)
    else:
        data_home = Path("~/.local/share").expanduser()
    return data_home / 'applications'


def create_desktop_file(path, target, display_name, icon_file, overwrite=True):
    path = Path(path)
    path.parent.mkdir(parents=True, exist_ok=True)
    if not overwrite and path.exists():
        raise FileExistsError(path)

    # TODO: customise declaring files accepted... %F arguments etc. Would require
    # additional config

    contents = f"""\
            [Desktop Entry]
            Name={display_name}
            Exec={_desktop_escape(target)}
            Icon={_desktop_escape(icon_file)}
            Type=Application
            """
    path.write_text(dedent(contents), encoding='utf8')



import os
from setuptools import setup, Extension
import platform

VERSION_SCHEME = {
    "version_scheme": os.getenv("SCM_VERSION_SCHEME", "guess-next-dev"),
    "local_scheme": os.getenv("SCM_LOCAL_SCHEME", "node-and-date"),
}

# The extension still exists on unix, it's just an empty module. This ensures wheel
# knows it is an impure package.
WINDOWS = platform.system() == 'Windows'
EXT_MODULES = [
    Extension(
        'desktop_app.wineventhook',
        sources=[os.path.join('src', 'wineventhook.cpp')],
        libraries=["user32", "shell32", "ole32"] if WINDOWS else [],
    )
]

setup(
    use_scm_version=VERSION_SCHEME,
    ext_modules=EXT_MODULES,
)
